"""
Custom Annotations for `charz-core`
===================================

This file contains private annotations used across this package.

Whenever there is a "?" comment,
it means a type may or may not implement that field or mixin class.
"""

from __future__ import annotations as _annotations

import sys as _sys
from itertools import count as _count
from typing import (
    TypeVar as _TypeVar,
    TypeAlias as _TypeAlias,
    ClassVar as _ClassVar,
    Hashable as _Hashable,
    Protocol as _Protocol,
    overload as _overload,
)

from linflex import Vec2 as _Vec2

if _sys.version_info >= (3, 11):
    from typing import (
        LiteralString as _LiteralString,
        Self,
    )
else:
    from typing_extensions import (
        LiteralString as _LiteralString,
        Self,
    )

from ._frame_task import FrameTaskManager as _FrameTaskManager

T = _TypeVar("T")
NodeID: _TypeAlias = int
GroupID: _TypeAlias = _LiteralString | NodeID | _Hashable


class Engine(_Protocol):
    frame_tasks: _FrameTaskManager[Self]  # Global across instances
    _is_running: bool

    @property
    def is_running(self) -> bool: ...
    @is_running.setter
    def is_running(self, run_state: bool) -> None: ...
    def process(self) -> None: ...
    def update(self) -> None: ...
    def run(self) -> None: ...


class Node(_Protocol):
    _uid_counter: _ClassVar[_count[NodeID]]
    uid: NodeID
    parent: Node | None

    def __init__(self) -> None: ...
    def with_parent(self, parent: Node | None, /) -> Self: ...
    def update(self) -> None: ...
    def queue_free(self) -> None: ...


class TransformComponent(_Protocol):
    position: _Vec2
    rotation: float
    top_level: bool

    @_overload
    def with_position(
        self,
        position: _Vec2,
        /,
    ) -> Self: ...
    @_overload
    def with_position(
        self,
        *,
        x: float,
        y: float,
    ) -> Self: ...
    @_overload
    def with_global_position(
        self,
        global_position: _Vec2,
        /,
    ) -> Self: ...
    @_overload
    def with_global_position(
        self,
        *,
        x: float,
        y: float,
    ) -> Self: ...
    def with_rotation(self, rotation: float, /) -> Self: ...
    def with_global_rotation(self, global_rotation: float, /) -> Self: ...
    def with_top_level(self, state: bool = True, /) -> Self: ...
    @property
    def global_position(self) -> _Vec2: ...
    @global_position.setter
    def global_position(self, position: _Vec2) -> None: ...
    @property
    def global_rotation(self) -> float: ...
    @global_rotation.setter
    def global_rotation(self, rotation: float) -> None: ...
    def set_global_x(self, x: float, /) -> None: ...
    def set_global_y(self, y: float, /) -> None: ...


class TransformNode(
    TransformComponent,
    Node,
    _Protocol,
): ...



from __future__ import annotations as _annotations

from linflex import Vec2

from .._node import Node
from .._components.transform import TransformComponent


class Node2D(TransformComponent, Node):
    """`Node2D` node that exists in 2D space.

    Has a transform (position, rotation).
    All 2D nodes, including sprites, inherit from Node2D.
    Use Node2D as a parent node to move, hide and rotate children in a 2D project.

    Example:

    Extending `Node2D` with components:

    ```python
    from charz_core import Node2D

    class ColorComponent:
        color: str = "green"

    class GreenPoint(Node2D, ColorComponent):
        ...

    assert ColoredPoint().color == "green"
    ```
    """

    def __init__(
        self,
        parent: Node | None = None,
        *,
        position: Vec2 | None = None,
        rotation: float | None = None,
        top_level: bool | None = None,
    ) -> None:
        Node.__init__(
            self,
            parent=parent,
        )
        if position is not None:
            self.position = position
        if rotation is not None:
            self.rotation = rotation
        if top_level is not None:
            self.top_level = top_level

    def __repr__(self) -> str:
        return (
            self.__class__.__name__
            + "("
            + f"#{self.uid}"
            + f":{round(self.position, 2)}"
            + f":{round(self.rotation, 2)}R"
            + ")"
        )



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, ClassVar, Any

from typing_extensions import Self

from ._frame_task import FrameTaskManager
from ._grouping import Group
from ._annotations import T, GroupID, NodeID

if TYPE_CHECKING:
    from ._node import Node


def static_load_node_type() -> type[Node]:
    """Workaround for the static type checker, to prevent circular dependencies."""
    if TYPE_CHECKING:
        from ._node import Node

        return Node
    return None  # Value used at runtime


# This type variable is never called, as its true value is `None` at runtime
# Used by the static type checker
NodeType = static_load_node_type()


class SceneClassProperties(type):
    _current: Scene

    @property
    def current(cls) -> Scene:
        if not hasattr(cls, "_current"):
            cls._current = cls()  # Create default scene if none exists
        return cls._current

    @current.setter
    def current(cls, new: Scene) -> None:
        cls.current.on_exit()
        cls._current = new
        new.on_enter()


class Scene(metaclass=SceneClassProperties):
    """`Scene` to encapsulate dimensions/worlds.

    Instantiating a scene (either of type `Scene` or subclass of `Scene`),
    will set that new scene instance to the current scene.

    Example:

    Structure of a scene and how to declare:

    ```python
    from charz_core import Scene
    from .my_files.player import Player

    class InsideHouse(Scene):
        def __init__(self) -> None:
            self.player = Player(position=Vec2(5, 7))  # Player when inside house
            self.table = ...
            self.chair = ...
    ```

    `NOTE` Use the *classmethod* `preload` to prevent setting current scene,
    while still loading nodes (and more) of the returned instance.

    When a node is created, it will be handled by the currently active `Scene`.

    `NOTE` If no `Scene` is created,
    a default `Scene` will be created and set as the active one.

    By subclassing `Scene`, and implementing `__init__`, all nodes
    created in that `__init__` will be added to that subclass's group of nodes.

    `NOTE (Technical)` A `Scene` hitting reference count of `0`
    will reduce the reference count to its nodes by `1`.
    """

    # Tasks are shared across all scenes
    frame_tasks: ClassVar[FrameTaskManager[Self]] = FrameTaskManager()
    # Values are set in `Scene.__new__`
    groups: defaultdict[GroupID, dict[NodeID, Node]]
    _queued_nodes: set[NodeID]

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        instance = super().__new__(cls, *args, **kwargs)
        # NOTE: When instantiating the scene,
        #       it will be set as the current one
        #     - Use preloading to surpass
        Scene._current = instance
        instance.groups = defaultdict(dict)
        instance._queued_nodes = set()
        return instance

    @classmethod
    def preload(cls) -> Self:
        """Preload the scene class, creating an instance without setting it as current.

        Returns:
            Self: An instance of the scene class, without setting it as current.
        """
        previous_scene = Scene.current
        instance = cls()
        Scene.current = previous_scene
        return instance

    def __init__(self) -> None:  # Override in subclass
        """Override to instantiate nodes and state related to this scene."""

    def __repr__(self) -> str:
        group_counts = ", ".join(f"{group}: {len(self.groups[group])}" for group in Group)
        return f"{self.__class__.__name__}({group_counts})"

    def set_current(self) -> None:
        """Set this scene as the current one."""
        Scene.current = self

    def as_current(self) -> Self:
        """Chained method to set this scene as the current one."""
        self.set_current()
        return self

    def get_group_members(
        self,
        group_id: GroupID,
        /,
        # Used to hint type checkers of dynamic return type
        type_hint: type[T] = NodeType,
    ) -> list[T]:
        """Get all members of a specific group.

        Args:
            group_id (GroupID): The ID of the group to retrieve members from.
            type_hint (type[T], optional): Node type in list returned.
                Defaults to type[Node].

        Returns:
            list[T]: A list of nodes in the specified group.
        """
        # NOTE: Return type `list` is faster than `tuple`,
        #       when copying iterate a copy (hence the use of `list(...)`)
        #       This allows node creation during iteration
        return list(self.groups[group_id].values())  # type: ignore

    def get_first_group_member(
        self,
        group_id: GroupID,
        /,
        # Used to hint type checkers of dynamic return type
        type_hint: type[T] = NodeType,
    ) -> T:
        """Get the first member of a specific group.

        Args:
            group_id (GroupID): The ID of the group to retrieve the first member from.
            type_hint (type[T], optional): Node type returned.
                Defaults to type[Node].

        Returns:
            T: The first node in the specified group.

        Raises:
            ValueError: If the group is empty.
        """
        for node in self.groups[group_id].values():
            return node  # type: ignore
        raise ValueError(f"No node in group {group_id}")

    def process(self) -> None:
        """Process the scene, executing all frame tasks.

        This method is called each frame to update the scene and its nodes,
        but can also be called manually to simulate time step.
        It will execute all registered frame tasks in the order of their priority.
        """
        for frame_task in self.frame_tasks.values():
            frame_task(self)

    def update(self) -> None:
        """Called each frame.

        Override this method in new subclass to implement custom update logic.
        """

    def on_enter(self) -> None:
        """Triggered when this scene is set as the current one."""

    def on_exit(self) -> None:
        """Triggered when this scene is no longer the current one."""


# Define core frame tasks


def update_self_scene(current_scene: Scene) -> None:
    """Update the scene itself, calling `update` on current scene."""
    current_scene.update()


def update_nodes(current_scene: Scene) -> None:
    """Update all nodes in the current scene, calling `update` on each node."""
    for node in current_scene.get_group_members(Group.NODE):
        node.update()


def free_queued_nodes(current_scene: Scene) -> None:
    """Free all queued nodes in the current scene, called at the end of each frame."""
    for node_id in current_scene._queued_nodes:
        for group in current_scene.groups.values():
            if node_id in group:
                del group[node_id]
    current_scene._queued_nodes.clear()


# Register core frame tasks
# Priorities are chosen with enough room to insert many more tasks in between
Scene.frame_tasks[100] = update_self_scene
Scene.frame_tasks[90] = update_nodes
Scene.frame_tasks[80] = free_queued_nodes



from __future__ import annotations

from enum import Flag, unique, auto
from typing import Literal

from linflex import Vec2
from typing_extensions import Self

from ._node import Node, NodeMixinSorter
from ._prefabs.node2d import Node2D


@unique
class CameraMode(Flag):
    FIXED = auto()
    CENTERED = auto()
    INCLUDE_SIZE = auto()


class CameraClassAttributes(NodeMixinSorter):
    """Workaround to add class attributes to `Camera`."""

    MODE_FIXED: CameraMode = CameraMode.FIXED
    MODE_CENTERED: CameraMode = CameraMode.CENTERED
    MODE_INCLUDE_SIZE: CameraMode = CameraMode.INCLUDE_SIZE
    _current: Camera

    @property
    def current(self) -> Camera:
        if not hasattr(self, "_current"):
            self._current = Camera()  # Create default camera if none exists
        return self._current

    @current.setter
    def current(self, new: Camera) -> None:
        self._current = new


class Camera(Node2D, metaclass=CameraClassAttributes):
    """`Camera` for controlling location of viewport in the world, per `Scene`.

    To access the current camera, access `Camera.current`.
    To set the current camera, use `Camera.current = <Camera>`,
    or call `camera.set_current()` on the camera instance.
    To set the current camera and return the instance, use `camera.as_current()`.

    `NOTE` A default `Camera` will be used if not explicitly set.

    Example:

    ```python
    from charz_core import Engine, Camera

    class MyGame(Engine):
        def __init__(self) -> None:
            # Configure how the current camera centers the viewport
            Camera.current.mode = Camera.MODE_CENTERED | Camera.MODE_INCLUDE_SIZE
    ```

    Attributes:
        `current`: `ClassVar[property[Camera]]` - Current camera instance in use.
        `mode`: `CameraMode` - Mode to decide origin for centering.

    Variants for `Camera.mode`:
    - `Camera.MODE_FIXED`: Camera is fixed at upper left corner.
    - `Camera.MODE_CENTERED`: Camera is centered.
    - `Camera.MODE_INCLUDE_SIZE`: Camera includes texture size of parent to camera.

    Methods:
        `set_current`
        `as_current` `chained`
        `is_current`
        `with_mode` `chained`
    """

    mode: CameraMode = CameraMode.FIXED

    def __init__(
        self,
        parent: Node | None = None,
        *,
        position: Vec2 | None = None,
        rotation: float | None = None,
        top_level: bool | None = None,
        mode: CameraMode | None = None,
        current: Literal[True] | None = None,
    ) -> None:
        Node2D.__init__(
            self,
            parent=parent,
            position=position,
            rotation=rotation,
            top_level=top_level,
        )
        if mode is not None:
            self.mode = mode
        if current is True:
            self.set_current()

    def set_current(self) -> None:
        """Set this camera as the current camera."""
        Camera.current = self

    def as_current(self) -> Self:
        """Chained method to set this camera as the current camera.

        Returns:
            Self: Same camera instance.
        """
        self.set_current()
        return self

    def is_current(self) -> bool:
        """Check if this camera is the current camera of the current `Scene`.

        Returns:
            bool: `True` if this camera is the current camera, `False` otherwise.
        """
        return Camera.current is self

    def with_mode(self, mode: CameraMode, /) -> Self:
        """Chained method to set the camera's mode.

        Args:
            mode (CameraMode): Enum variant to set the camera's mode.
        """
        self.mode = mode
        return self



from __future__ import annotations

from itertools import count
from typing import Any, ClassVar

from typing_extensions import Self

from ._grouping import Group, group
from ._scene import Scene
from ._annotations import NodeID


class NodeMixinSorter(type):
    """Metaclass for sorting `Node` to be the last base in MRO."""

    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        attrs: dict[str, object],
    ) -> type:
        def sorter(base: type) -> bool:
            # Ensures `Node` will be last in MRO (except for `object`),
            # to help `__new__` chain for nodes (instances of `Node`/`Node` subclass)
            # work despite wrong order in subclass declaration
            return issubclass(base, Node)

        sorted_bases = tuple(sorted(bases, key=sorter))
        new_type = super().__new__(cls, name, sorted_bases, attrs)
        return new_type


@group(Group.NODE)
class Node(metaclass=NodeMixinSorter):
    """`Node` base class.

    All nodes existing in a scene (either "physically" or "theoretically"),
    will be instances of either this class, or a subclass.
    Subclasses can be combined with component mixins, using Python's multiple inheritance.
    This class can be instantiated directly, but in itself it does not do anything useful.

    The reference of nodes is stored in `Scene.groups[Group.NODE]`,
    which is a dictionary mapping `NodeID` to `Node` instances.
    It keeps the node alive, and prevents it from being garbage collected.

    Example:

    Every node will be assigned a unique identifier (`uid`),
    which can be used to reference it within the scene:

    ```python
    from charz_core import Scene, Node, Group

    my_node = Node()
    reference = Scene.groups[Group.NODE][my_node.uid]
    assert reference is my_node
    ```

    The assignment of `uid` is done automatically when the node is created,
    and is guaranteed to be unique within the scene.
    It happens at the end of the `__new__` chain.
    The `__new__` chain is when components call `super().__new__(...)`,
    which calls `Node.__new__` to assign the `uid`, since the bases are
    sorted to ensure `Node` is the last base in the MRO (Method Resolution Order).
    """

    _uid_counter: ClassVar[count[NodeID]] = count(0, 1)

    def __new__(cls, *_args: Any, **_kwargs: Any) -> Self:
        """Last in the `__new__` chain, which assigns `uid`."""
        # NOTE: Additional args and kwargs are ignored!
        instance = super().__new__(cls)
        instance.uid = next(Node._uid_counter)
        return instance

    uid: NodeID  # Is set in `Node.__new__`
    parent: Node | None = None

    def __init__(self, parent: Node | None = None) -> None:
        """Initialize node."""
        if parent is not None:
            self.parent = parent

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(#{self.uid})"

    def with_parent(self, parent: Node | None, /) -> Self:
        """Chained method to set parent node."""
        self.parent = parent
        return self

    def update(self) -> None:
        """Called each frame.

        Override this method in subclasses to implement custom update logic.
        """

    def queue_free(self) -> None:
        """Queues this node for freeing.

        This method should be called when you want to remove the node from the scene.
        It will be freed at the end of the current frame, handled by `Engine.frame_tasks`.
        """
        Scene.current._queued_nodes.add(self.uid)



"""
Charz Core
==========

Core logic for `charz`

Includes
--------

- Annotations
  - `Self`  (from standard `typing` or from package `typing-extensions`)
- Math (from package `linflex`)
  - `lerp`
  - `sign`
  - `clamp`
  - `move_toward`
  - `Vec2`
  - `Vec2i`
  - `Vec3`
- Framework
  - `Engine`
  - `Scene`
- Decorators
  - `group`
- Enums
  - `Group`
- Components
  - `TransformComponent`
- Nodes
  - `Camera`
  - `Node`
  - `Node2D`
"""

__all__ = [
    "Engine",
    "Camera",
    "Scene",
    "group",
    "Group",
    "Node",
    "Self",
    "Node2D",
    "TransformComponent",
    "lerp",
    "sign",
    "clamp",
    "move_toward",
    "Vec2",
    "Vec2i",
    "Vec3",
]

# Re-exports
from linflex import lerp, sign, clamp, move_toward, Vec2, Vec2i, Vec3

# Exports
from ._annotations import Self  # Version proof
from ._engine import Engine
from ._camera import Camera
from ._scene import Scene
from ._grouping import Group, group
from ._node import Node
from ._components.transform import TransformComponent
from ._prefabs.node2d import Node2D



from __future__ import annotations

from typing import ClassVar

from typing_extensions import Self

from ._frame_task import FrameTaskManager
from ._scene import Scene


class EngineMixinSorter(type):
    """Metaclass for sorting `Engine` to be the last base in MRO."""

    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        attrs: dict[str, object],
    ) -> type:
        def sorter(base: type) -> bool:
            # Ensures `Engine` will be last in MRO (except for `object`),
            # to help `__new__` chain for nodes (instances of `Engine`/`Engine` subclass)
            # work despite wrong order in subclass declaration
            return issubclass(base, Engine)

        sorted_bases = tuple(sorted(bases, key=sorter))
        new_type = super().__new__(cls, name, sorted_bases, attrs)
        return new_type


class Engine(metaclass=EngineMixinSorter):
    """`Engine` for managing the game loop and frame tasks.

    Subclass this to create your main entry class.

    Example:

    ```python
    class MyGame(Engine):
        def __init__(self) -> None:
            ... # Initialize your nodes, preload scenes, etc.

        def update(self) -> None:
            ... # Your game logic here
    ```
    """

    # Tasks are shared across all engines
    frame_tasks: ClassVar[FrameTaskManager[Self]] = FrameTaskManager()
    # Using setter and getter to prevent subclass def overriding
    _is_running: bool = False

    @property
    def is_running(self) -> bool:
        """Check if main loop is running.

        This attribute is wrapped in a property to protect it from being
        overridden by subclass definitions.

        Example:

        This will signal the type checker the following is not allowed:

        ```python
        from charz_core import Engine

        class MyGame(Engine):
            is_running: bool = True  # Invalid type, reported by type checker
        ```

        Returns:
            bool: `True` if the main loop is running, `False` otherwise.
        """
        return self._is_running

    @is_running.setter
    def is_running(self, run_state: bool) -> None:
        """Set the running state of the main loop.

        Args:
            run_state (bool): `True` to start/continue the main loop, `False` to stop it.
        """
        self._is_running = run_state

    def update(self) -> None:
        """Called each frame.

        Override this method in new subclass to implement custom update logic.
        """

    def run(self) -> None:
        """Run app/game, which will start the main loop.

        This will run the main loop until `is_running` is set to `False`.
        """
        self.is_running = True
        while self.is_running:
            for frame_task in self.frame_tasks.values():
                frame_task(self)


# Define core frame tasks


def update_self_engine(engine: Engine) -> None:
    """Update the engine itself, calling its `update` method."""
    engine.update()


def process_current_scene(_engine: Engine) -> None:
    """Process the current scene (`Scene.current`), calling its `process` method."""
    Scene.current.process()


# Register core frame tasks
# Priorities are chosen with enough room to insert many more tasks in between
Engine.frame_tasks[100] = update_self_engine
Engine.frame_tasks[90] = process_current_scene



from __future__ import annotations

from typing import Any

from linflex import Vec2
from typing_extensions import Self
from typing import overload


class TransformComponent:  # Component (mixin class)
    """`TransformComponent` mixin class for node.

    Examples:

    Composing a 2D node with custom physics component:

    ```python
    from charz_core import Node, TransformComponent
    from .my_files.physics_component import PhysicsComponent

    class PhysicsBody(TransformComponent, PhysicsComponent, Node):
        ...
    ```

    *Psudocode* for how `charz_core.Node2D` is composed:

    ```python
    from charz_core import Node, TransformComponent

    class Node2D(TransformComponent, Node):
        ...
    ```

    Attributes:
        `position`: `Vec2` - Position in local space.
        `rotation`: `float` - Angle in radians.
        `top_level`: `bool` - Indicating if the node is a top-level node.
        `global_position`: `property[Vec2]` - Copy of position in world space.
        `global_rotation`: `property[float]` - Rotation in world space.

    Methods:
        `set_global_x`
        `set_global_y`
    """

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        instance = super().__new__(cls, *args, **kwargs)
        if (class_position := getattr(instance, "position", None)) is not None:
            instance.position = Vec2.copy(class_position)
        else:
            instance.position = Vec2.ZERO
        return instance

    position: Vec2
    rotation: float = 0
    top_level: bool = False

    @overload
    def with_position(
        self,
        position: Vec2,
        /,
    ) -> Self: ...

    @overload
    def with_position(
        self,
        *,
        x: float,
        y: float,
    ) -> Self: ...

    def with_position(  # type: ignore[override]
        self,
        position: Vec2 | None = None,
        /,
        *,
        x: float | None = None,
        y: float | None = None,
    ) -> Self:
        """Chained method to set the node's position.

        This method allows you to set the position of the node,
        using either a `Vec2` instance or individual `x` and `y` coordinates.

        Args:
            position (Vec2 | None, optional): Position of the node. Defaults to None.
            x (float | None, optional): X-coordinate of the node. Defaults to None.
            y (float | None, optional): Y-coordinate of the node. Defaults to None.

        Raises:
            TypeError: If all arguments are `None` at the same time.
            TypeError: If both `position` and any of `x`/`y` are provided.

        Returns:
            Self: Same node instance.
        """
        if position is None and x is None and y is None:
            raise TypeError(f"Not all arguments can be {None} at the same time")
        if position is not None and (x is not None or y is not None):
            raise TypeError(
                "Chose either positional argument 'position' "
                "or keyword arguments 'x' and/or 'y', not all three"
            )
        if position is not None:
            self.position = position
        if x is not None:
            self.position.x = x
        if y is not None:
            self.position.y = y
        return self

    @overload
    def with_global_position(
        self,
        global_position: Vec2,
        /,
    ) -> Self: ...

    @overload
    def with_global_position(
        self,
        *,
        x: float,
        y: float,
    ) -> Self: ...

    def with_global_position(  # type: ignore[override]
        self,
        global_position: Vec2 | None = None,
        /,
        *,
        x: float | None = None,
        y: float | None = None,
    ) -> Self:
        """Chained method to set the node's global position.

        This method allows you to set the global position of the node,
        using either a `Vec2` instance or individual `x` and `y` coordinates.

        Args:
            global_position (Vec2 | None, optional): Global position. Defaults to None.
            x (float | None, optional): Global x-coordinate of node. Defaults to None.
            y (float | None, optional): Global y-coordinate of node. Defaults to None.

        Raises:
            TypeError: If all arguments are `None` at the same time.
            TypeError: If both `global_position` and any of `x`/`y` are provided.

        Returns:
            Self: Same node instance.
        """
        if global_position is None and x is None and y is None:
            raise TypeError(f"Not all arguments can be {None} at the same time")
        if global_position is not None and (x is not None or y is not None):
            raise TypeError(
                "Chose either positional argument 'global_position' "
                "or keyword arguments 'x' and/or 'y', not all three"
            )
        if global_position is not None:
            self.global_position = global_position
        if x is not None:
            self.set_global_x(x)
        if y is not None:
            self.set_global_y(y)
        return self

    def with_rotation(self, rotation: float, /) -> Self:
        """Chained method to set the node's `rotation`.

        Args:
            rotation (float): Rotation in radians.

        Returns:
            Self: Same node instance.
        """
        self.rotation = rotation
        return self

    def with_global_rotation(self, global_rotation: float, /) -> Self:
        """Chained method to set the node's `global_rotation`.

        Args:
            global_rotation (float): Global rotation in radians.

        Returns:
            Self: Same node instance.
        """
        self.global_rotation = global_rotation
        return self

    def with_top_level(self, state: bool = True, /) -> Self:
        """Chained method to set the node's `top_level` state.

        Args:
            state (bool, optional): Whether node is a top-level node. Defaults to True.

        Returns:
            Self: Same node instance.
        """
        self.top_level = state
        return self

    def set_global_x(self, x: float, /) -> None:
        """Set node's global x-coordinate.

        Args:
            x (float): Global x-coordinate.
        """
        diff_x = x - self.global_position.x
        self.position.x += diff_x

    def set_global_y(self, y: float, /) -> None:
        """Set node's global y-coordinate.

        Args:
            y (float): Global y-coordinate.
        """
        diff_y = y - self.global_position.y
        self.position.y += diff_y

    @property
    def global_position(self) -> Vec2:
        """Computes a copy of the node's global position (world space).

        `NOTE` Cannot do the following:

        ```python
        self.global_position.x += 5
        self.global_position.x = 42
        ```

        Instead, you should use:

        ```python
        self.position.x += 5
        self.set_global_x(42)
        ```

        Returns:
            Vec2: Copy of global position.
        """
        if self.top_level:
            return self.position.copy()
        global_position = self.position.copy()
        parent = self.parent  # type: ignore
        while isinstance(parent, TransformComponent):
            # Check for rotation, since cos(0) and sin(0) produces *approximate* values
            if parent.rotation:
                global_position = parent.position + global_position.rotated(
                    parent.rotation
                )
            else:
                global_position += parent.position
            if parent.top_level:
                return global_position
            parent = parent.parent  # type: ignore
        return global_position

    @global_position.setter
    def global_position(self, position: Vec2) -> None:
        """Set node's global position (world space).

        Args:
            position (Vec2): Global position.
        """
        diff = position - self.global_position
        self.position += diff

    @property
    def global_rotation(self) -> float:
        """Computes node's global rotation (world space).

        Returns:
            float: Global rotation in radians.
        """
        if self.top_level:
            return self.rotation
        global_rotation = self.rotation
        parent = self.parent  # type: ignore
        while isinstance(parent, TransformComponent):
            global_rotation += parent.rotation
            if parent.top_level:
                return global_rotation
            parent = parent.parent  # type: ignore
        return global_rotation

    @global_rotation.setter
    def global_rotation(self, rotation: float) -> None:
        """Set node's global rotation (world space).

        Args:
            rotation (float): Global rotation in radians.
        """
        diff = rotation - self.global_rotation
        self.rotation += diff



from __future__ import annotations

import sys
from enum import unique
from functools import wraps
from typing import TYPE_CHECKING, Callable, Any

from ._annotations import T, GroupID

if TYPE_CHECKING:
    from ._node import Node


# NOTE: Variants of the enum produces the same hash as if it was using normal `str`
if sys.version_info >= (3, 11):
    from enum import StrEnum, auto

    @unique
    class Group(StrEnum):
        """Enum for core node groups used in `charz-core`."""

        NODE = auto()

else:
    from enum import Enum

    @unique
    class Group(str, Enum):
        """Enum for core node groups used in `charz-core`."""

        NODE = "node"


def group(group_id: GroupID, /) -> Callable[[type[T]], type[T]]:
    """Decorator that adds `node`/`component` to the given `group`.

    Example:

    Adding instances of a custom node class to a new group `"tile"`,
    and `iterating over members` in the current scene:

    ```python
    from charz_core import Node2D, Scene, group

    @group("tile")
    class WorldTile(Node2D):
        def __init__(self, material_name: str) -> None:
            self.material_name = material_name

    dirt = WorldTile("Dirt")
    stone = WorldTile("Stone")

    for tile in Scene.current.get_group_members("tile", type_hint=WorldTile):
        print(tile.material_name)

    # Prints out
    >>> 'Dirt'
    >>> 'Stone'
    ```

    This works by wrapping `__new__` and `_free`.
    Recommended types for parameter `group_id`: `LiteralString`, `StrEnum` or `int`.

    `NOTE` Each node is added to the current scene's group when `__new__` is called.

    Args:
        group_id (GroupID): *Hashable* object used for group ID

    Returns:
        Callable[[type[T]], type[T]]: Wrapped class
    """
    # NOTE: Lazyloading `Scene`
    # Do import here to prevent cycling dependencies,
    # as there won't be a lot of scene creation
    from ._scene import Scene

    def wrapper(node_type_or_component: type[T]) -> type[T]:
        original_new = node_type_or_component.__new__

        # This will not always be correct, but I will leave it for now...
        @wraps(original_new)
        def new_wrapper(cls: type[T], *args: Any, **kwargs: Any) -> T:
            # This conditional fixes (hopefully) the MRO chain problem I had for a long time...
            if original_new is object.__new__:
                # Perform `super()` call to the next object's `__new__` in the MRO chain
                # without the help of compiler magic.
                # This path (in the if-statement) is likely triggered by
                # the component/node class (that is decorated with this decorator)
                # not implementing its own `__new__` method.
                # NOTE: This path fixes components' `__new__` so it will call
                # the right `__new__` associated with the next object in the MRO chain,
                # and don't just end with `object.__new__`
                # - that previously resulted in "lost logic".
                mro_next_index = cls.__mro__.index(node_type_or_component) + 1
                mro_next = cls.__mro__[mro_next_index]
                instance = mro_next.__new__(cls, *args, **kwargs)
            else:
                # Wrap the original `__new__` method
                instance = original_new(cls, *args, **kwargs)
            Scene.current.groups[group_id][instance.uid] = instance  # type: ignore
            return instance

        node_type_or_component.__new__ = new_wrapper
        return node_type_or_component

    return wrapper



from __future__ import annotations

from typing import TypeVar, TypeAlias, Generic, Callable

# NOTE: These annotaion-related variables exist here to prevent circular import error
T = TypeVar("T")
Priority: TypeAlias = int
FrameTask: TypeAlias = Callable[[T], None]


class FrameTaskManager(Generic[T], dict[Priority, FrameTask[T]]):
    """A dict-like manager that auto-sorts tasks by priority.

    `NOTE` The higher the priority, the earlier the task will be executed.
    """

    def __setitem__(self, key: Priority, value: FrameTask[T]) -> None:
        super().__setitem__(key, value)
        sorted_items = sorted(self.items(), reverse=True)
        self.clear()
        self.update(sorted_items)

    def __repr__(self) -> str:
        # Produces: [Priority]: [Function Name] :: [Function Docstring?]
        return (
            self.__class__.__name__
            + f"[{T}]("
            + "".join(
                f"\n| {priority}: {function.__qualname__} -- "
                f'"{"<No Docstring>" if function.__doc__ is None else function.__doc__}"'
                for (priority, function) in self.items()
            )
            + "\n)"
        )
